"""Test for the generic geojson feed entry."""
from tests import MockSimpleFeedEntry


def test_simple_feed_entry():
    """Test feed entry behaviour."""
    feed_entry = MockSimpleFeedEntry(None, None)
    assert repr(feed_entry) == "<MockSimpleFeedEntry(id=mock id)>"
    assert feed_entry.geometries is None
    assert feed_entry.coordinates is None
    assert feed_entry.title == "mock title"
    assert feed_entry.external_id == "mock id"
    assert feed_entry.attribution is None



"""GeoJSON distance helper."""
from __future__ import annotations

import logging

from haversine import haversine

from .geometries import Geometry, Point, Polygon

_LOGGER = logging.getLogger(__name__)


class GeoJsonDistanceHelper:
    """Helper to calculate distances between GeoJSON geometries."""

    def __init__(self):
        """Initialise the geo distance helper."""
        pass

    @staticmethod
    def extract_coordinates(geometry: Geometry) -> tuple[float, float] | None:
        """Extract the best coordinates from the feature for display."""
        latitude = longitude = None
        if isinstance(geometry, Point):
            # Just extract latitude and longitude directly.
            latitude, longitude = geometry.latitude, geometry.longitude
        elif isinstance(geometry, Polygon):
            centroid = geometry.centroid
            latitude, longitude = centroid.latitude, centroid.longitude
            _LOGGER.debug("Centroid of %s is %s", geometry, (latitude, longitude))
        else:
            _LOGGER.debug("Not implemented: %s", type(geometry))
        return latitude, longitude

    @staticmethod
    def distance_to_geometry(
        coordinates: tuple[float, float], geometry: Geometry
    ) -> float:
        """Calculate the distance between coordinates and geometry."""
        distance = float("inf")
        if isinstance(geometry, Point):
            distance = GeoJsonDistanceHelper._distance_to_point(coordinates, geometry)
        elif isinstance(geometry, Polygon):
            distance = GeoJsonDistanceHelper._distance_to_polygon(coordinates, geometry)
        else:
            _LOGGER.debug("Not implemented: %s", type(geometry))
        return distance

    @staticmethod
    def _distance_to_point(coordinates: tuple[float, float], point: Point) -> float:
        """Calculate the distance between coordinates and the point."""
        # Swap coordinates to match: (latitude, longitude).
        return GeoJsonDistanceHelper._distance_to_coordinates(
            coordinates, (point.latitude, point.longitude)
        )

    @staticmethod
    def _distance_to_polygon(
        coordinates: tuple[float, float], polygon: Polygon
    ) -> float:
        """Calculate the distance between coordinates and the polygon."""
        distance = float("inf")
        # Check if coordinates are inside the polygon.
        if polygon.is_inside(Point(coordinates[0], coordinates[1])):
            return 0.0
        # Calculate distance from polygon by calculating the distance
        # to each point of the polygon.
        for polygon_point in polygon.points:
            distance = min(
                distance,
                GeoJsonDistanceHelper._distance_to_point(coordinates, polygon_point),
            )
        # Next calculate the distance to each edge of the polygon.
        for edge in polygon.edges:
            distance = min(
                distance, GeoJsonDistanceHelper._distance_to_edge(coordinates, edge)
            )
        _LOGGER.debug("Distance between %s and %s: %s", coordinates, polygon, distance)
        return distance

    @staticmethod
    def _distance_to_coordinates(
        coordinates1: tuple[float, float], coordinates2: tuple[float, float]
    ) -> float:
        """Calculate the distance between two coordinates tuples.."""
        # Expecting coordinates in format: (latitude, longitude).
        return haversine(coordinates2, coordinates1)

    @staticmethod
    def _distance_to_edge(
        coordinates: tuple[float, float], edge: tuple[Point, Point]
    ) -> float:
        """Calculate distance between coordinates and provided edge."""
        perpendicular_point = GeoJsonDistanceHelper._perpendicular_point(
            edge, Point(coordinates[0], coordinates[1])
        )
        # If there is a perpendicular point on the edge -> calculate distance.
        # If there isn't, then the distance to the end points of the edge will
        # need to be considered separately.
        if perpendicular_point:
            distance = GeoJsonDistanceHelper._distance_to_point(
                coordinates, perpendicular_point
            )
            _LOGGER.debug("Distance between %s and %s: %s", coordinates, edge, distance)
            return distance
        return float("inf")

    @staticmethod
    def _perpendicular_point(edge: tuple[Point, Point], point: Point) -> Point | None:
        """Find a perpendicular point on the edge to the provided point."""
        a, b = edge
        # Safety check: a and b can't be an edge if they are the same point.
        if a == b:
            return None
        px = point.longitude
        py = point.latitude
        ax = a.longitude
        ay = a.latitude
        bx = b.longitude
        by = b.latitude
        # Alter longitude to cater for 180 degree crossings.
        if px < 0:
            px += 360.0
        if ax < 0:
            ax += 360.0
        if bx < 0:
            bx += 360.0
        if ay > by or ax > bx:
            ax, ay, bx, by = bx, by, ax, ay
        dx = abs(bx - ax)
        dy = abs(by - ay)
        shortest_length = ((dx * (px - ax)) + (dy * (py - ay))) / (
            (dx * dx) + (dy * dy)
        )
        rx = ax + dx * shortest_length
        ry = ay + dy * shortest_length
        if bx >= rx >= ax and by >= ry >= ay:
            if rx > 180:
                # Correct longitude.
                rx -= 360.0
            return Point(ry, rx)
        return None



"""Feed Entry."""
from __future__ import annotations

import logging
from abc import ABC, abstractmethod

import geojson
from geojson import Feature

from .geojson_distance_helper import GeoJsonDistanceHelper
from .geometries import Geometry, Point, Polygon

_LOGGER = logging.getLogger(__name__)


class FeedEntry(ABC):
    """Feed entry base class."""

    def __init__(self, home_coordinates: tuple[float, float], feature: Feature):
        """Initialise this feed entry."""
        self._home_coordinates = home_coordinates
        self._feature = feature

    def __repr__(self):
        """Return string representation of this entry."""
        return f"<{self.__class__.__name__}(id={self.external_id})>"

    @property
    def geometries(self) -> list[Geometry] | None:
        """Return all geometry details of this entry."""
        if self._feature:
            return FeedEntry._wrap(self._feature.geometry)
        return None

    @staticmethod
    def _wrap(geometry: geojson.geometry.Geometry) -> list[Geometry] | None:
        """Wrap data of the provided GeoJSON geometry."""
        if isinstance(geometry, geojson.geometry.Point):
            return [Point(geometry.coordinates[1], geometry.coordinates[0])]
        elif isinstance(geometry, geojson.geometry.GeometryCollection):
            result = []
            for entry in geometry.geometries:
                wrapped_geometry = FeedEntry._wrap(entry)
                if wrapped_geometry:
                    result += wrapped_geometry
            return result
        elif isinstance(geometry, geojson.geometry.Polygon):
            # Currently only support polygons without a hole
            # (https://tools.ietf.org/html/rfc7946#page-23).
            return [
                Polygon(
                    [
                        Point(coordinate[1], coordinate[0])
                        for coordinate in geometry.coordinates[0]
                    ]
                )
            ]
        else:
            _LOGGER.debug("Not implemented: %s", type(geometry))
            return None

    @property
    def coordinates(self) -> tuple[float, float] | None:
        """Return the best coordinates (latitude, longitude) of this entry."""
        # This looks for the first point in the list of geometries. If there
        # is no point then return the first entry.
        if self.geometries and len(self.geometries) >= 1:
            for entry in self.geometries:
                if isinstance(entry, Point):
                    return GeoJsonDistanceHelper.extract_coordinates(entry)
            # No point found.
            return GeoJsonDistanceHelper.extract_coordinates(self.geometries[0])
        return None

    @property
    @abstractmethod
    def title(self) -> str | None:
        """Return the title of this entry."""
        return None

    @property
    @abstractmethod
    def external_id(self) -> str | None:
        """Return the external id of this entry."""
        return None

    @property
    def attribution(self) -> str | None:
        """Return the attribution of this entry."""
        return None

    @property
    def distance_to_home(self) -> float:
        """Return the distance in km of this entry to the home coordinates."""
        # This goes through all geometries and reports back the closest
        # distance to any of them.
        distance = float("inf")
        if self.geometries and len(self.geometries) >= 1:
            for geometry in self.geometries:
                distance = min(
                    distance,
                    GeoJsonDistanceHelper.distance_to_geometry(
                        self._home_coordinates, geometry
                    ),
                )
        return distance

    def _search_in_feature(self, name):
        """Find an attribute in the feature object."""
        if self._feature and name in self._feature:
            return self._feature[name]
        return None

    def _search_in_properties(self, name):
        """Find an attribute in the feed entry's properties."""
        if (
            self._feature
            and self._feature.properties
            and name in self._feature.properties
        ):
            return self._feature.properties[name]
        return None



"""Tests for aio-geojson-client library."""
from __future__ import annotations

from datetime import datetime
from typing import Optional

from aio_geojson_client.feed import GeoJsonFeed
from aio_geojson_client.feed_entry import FeedEntry

ATTR_GUID = "guid"
ATTR_ID = "id"
ATTR_TITLE = "title"


class MockGeoJsonFeed(GeoJsonFeed):
    """Mock geojson feed."""

    def _extract_from_feed(self, feed) -> Optional:
        return None

    def _extract_last_timestamp(self, feed_entries) -> datetime | None:
        return None

    def _new_entry(self, home_coordinates, feature, global_data):
        """Generate a new entry."""
        return MockFeedEntry(home_coordinates, feature)


class MockSimpleFeedEntry(FeedEntry):
    """Mock feed entry."""

    @property
    def title(self) -> str | None:
        """Return title."""
        return "mock title"

    @property
    def external_id(self) -> str | None:
        """Return external id."""
        return "mock id"


class MockFeedEntry(FeedEntry):
    """Generic feed entry."""

    def __init__(self, home_coordinates, feature):
        """Initialise this service."""
        super().__init__(home_coordinates, feature)

    @property
    def title(self) -> str:
        """Return the title of this entry."""
        return self._search_in_properties(ATTR_TITLE)

    @property
    def external_id(self) -> str:
        """Return the external id of this entry."""
        """Find a suitable ID for the provided entry."""
        external_id = self._search_in_feature(ATTR_ID)
        if not external_id:
            external_id = self._search_in_properties(ATTR_ID)
        if not external_id:
            external_id = self._search_in_properties(ATTR_GUID)
        if not external_id:
            external_id = self.title
        if not external_id:
            # Use geometry as ID as a fallback.
            external_id = hash(self.coordinates)
        return external_id



"""Configuration for tests."""
import pytest
from aioresponses import aioresponses


@pytest.fixture
def mock_aioresponse():
    """Return aioresponse fixture."""
    with aioresponses() as m:
        yield m



"""GeoJSON polygon."""
from __future__ import annotations

from .geometry import Geometry
from .point import Point


class Polygon(Geometry):
    """Represents a polygon."""

    def __init__(self, points: list[Point]):
        """Initialise polygon."""
        self._points = points

    def __repr__(self):
        """Return string representation of this polygon."""
        return f"<{self.__class__.__name__}(centroid={self.centroid})>"

    def __hash__(self) -> int:
        """Return unique hash of this polygon."""
        return hash(self.points)

    def __eq__(self, other: object) -> bool:
        """Return if this object is equal to other object."""
        return self.__class__ == other.__class__ and self.points == other.points

    @property
    def points(self) -> list | None:
        """Return the points of this polygon."""
        return self._points

    @property
    def edges(self) -> list[tuple[Point, Point]]:
        """Return all edges of this polygon."""
        edges = []
        for i in range(1, len(self.points)):
            previous = self.points[i - 1]
            current = self.points[i]
            edges.append((previous, current))
        return edges

    @property
    def centroid(self) -> Point:
        """Find the polygon's centroid as a best approximation."""
        longitudes_list = [point.longitude for point in self.points]
        latitudes_list = [point.latitude for point in self.points]
        number_of_points = len(self.points)
        longitude = sum(longitudes_list) / number_of_points
        latitude = sum(latitudes_list) / number_of_points
        return Point(latitude, longitude)

    def is_inside(self, point: Point) -> bool:
        """Check if the provided point is inside this polygon."""
        if point:
            crossings = 0
            for edge in self.edges:
                if Polygon._ray_crosses_segment(point, edge):
                    crossings += 1
            return crossings % 2 == 1
        return False

    @staticmethod
    def _ray_crosses_segment(point: Point, edge: tuple[Point, Point]):
        """Use ray-casting algorithm to check provided point and edge."""
        a, b = edge
        px = point.longitude
        py = point.latitude
        ax = a.longitude
        ay = a.latitude
        bx = b.longitude
        by = b.latitude
        if ay > by:
            ax = b.longitude
            ay = b.latitude
            bx = a.longitude
            by = a.latitude
        # Alter longitude to cater for 180 degree crossings.
        if px < 0:
            px += 360.0
        if ax < 0:
            ax += 360.0
        if bx < 0:
            bx += 360.0

        if py == ay or py == by:
            py += 0.00000001
        if (py > by or py < ay) or (px > max(ax, bx)):
            return False
        if px < min(ax, bx):
            return True

        red = ((by - ay) / (bx - ax)) if (ax != bx) else float("inf")
        blue = ((py - ay) / (px - ax)) if (ax != px) else float("inf")
        return blue >= red



"""Test geometries."""
import unittest

from aio_geojson_client.geometries import Point, Polygon


class TestGeometries(unittest.TestCase):
    """Test geometries."""

    def test_point(self):
        """Test point."""
        point = Point(-37.1234, 149.2345)
        assert point.latitude == -37.1234
        assert point.longitude == 149.2345
        assert repr(point) == "<Point(latitude=-37.1234, longitude=149.2345)>"

    def test_point_equality(self):
        """Test points."""
        point1 = Point(10.0, 15.0)
        point2 = Point(10.0, 15.0)
        assert point1 == point2

    def test_polygon(self):
        """Test polygon."""
        polygon = Polygon(
            [
                Point(-30.1, 150.1),
                Point(-30.2, 150.2),
                Point(-30.4, 150.4),
                Point(-30.8, 150.8),
                Point(-30.1, 150.1),
            ]
        )
        assert len(polygon.points) == 5
        assert polygon.centroid.latitude == -30.32
        assert polygon.centroid.longitude == 150.32
        assert (
            repr(polygon) == "<Polygon(centroid="
            "<Point(latitude=-30.32, longitude=150.32)>)>"
        )

    def test_polygon_equality(self):
        """Test points."""
        polygon1 = Polygon(
            [
                Point(30.0, 30.0),
                Point(30.0, 35.0),
                Point(35.0, 35.0),
                Point(35.0, 30.0),
                Point(30.0, 30.0),
            ]
        )
        polygon2 = Polygon(
            [
                Point(30.0, 30.0),
                Point(30.0, 35.0),
                Point(35.0, 35.0),
                Point(35.0, 30.0),
                Point(30.0, 30.0),
            ]
        )
        assert polygon1 == polygon2

    def test_point_in_polygon_1(self):
        """Test if point is in polygon."""
        polygon = Polygon(
            [Point(30.0, 30.0), Point(30.0, 35.0), Point(35.0, 35.0), Point(30.0, 30.0)]
        )
        # 1. Outside
        assert not polygon.is_inside(Point(20.0, 20.0))
        # 2. Inside
        assert polygon.is_inside(Point(31.0, 32.0))
        # 3. Inside
        assert polygon.is_inside(Point(30.0, 32.0))
        # 4. Inside
        assert polygon.is_inside(Point(30.0, 35.0))
        # 5. Outside
        assert not polygon.is_inside(Point(34.0, 31.0))

    def test_point_in_polygon_2(self):
        """Test if point is in polygon."""
        polygon = Polygon(
            [
                Point(30.0, -30.0),
                Point(30.0, -25.0),
                Point(35.0, -25.0),
                Point(30.0, -30.0),
            ]
        )
        # 1. Outside
        assert not polygon.is_inside(Point(20.0, -40.0))
        # 2. Inside
        assert polygon.is_inside(Point(31.0, -28.0))
        # 3. Inside
        assert polygon.is_inside(Point(30.0, -28.0))
        # 4. Inside
        assert polygon.is_inside(Point(30.0, -25.0))
        # 5. Outside
        assert not polygon.is_inside(Point(34.0, -29.0))
        # 6. Invalid point
        assert not polygon.is_inside(None)



"""Test for the generic geojson feed manager."""
import asyncio
from http import HTTPStatus
from unittest import mock as async_mock

import aiohttp
import pytest

from aio_geojson_client.consts import UPDATE_OK_NO_DATA
from aio_geojson_client.feed_manager import FeedManagerBase
from aio_geojson_client.filter_definition import GeoJsonFeedFilterDefinition
from tests import MockGeoJsonFeed
from tests.utils import load_fixture


@pytest.mark.asyncio
async def test_feed_manager(mock_aioresponse):
    """Test the feed manager."""
    home_coordinates = (-31.0, 151.0)
    mock_aioresponse.get(
        "http://test.url/testpath",
        status=HTTPStatus.OK,
        body=load_fixture("generic_feed_1.json"),
    )

    async with aiohttp.ClientSession(loop=asyncio.get_running_loop()) as websession:
        feed = MockGeoJsonFeed(websession, home_coordinates, "http://test.url/testpath")

        # This will just record calls and keep track of external ids.
        generated_entity_external_ids = []
        updated_entity_external_ids = []
        removed_entity_external_ids = []

        async def _generate_entity(external_id):
            """Generate new entity."""
            generated_entity_external_ids.append(external_id)

        async def _update_entity(external_id):
            """Update entity."""
            updated_entity_external_ids.append(external_id)

        async def _remove_entity(external_id):
            """Remove entity."""
            removed_entity_external_ids.append(external_id)

        feed_manager = FeedManagerBase(
            feed, _generate_entity, _update_entity, _remove_entity
        )
        assert (
            repr(feed_manager) == "<FeedManagerBase(feed=<"
            "MockGeoJsonFeed(home=(-31.0, 151.0), "
            "url=http://test.url/testpath, "
            "radius=None)>)>"
        )
        await feed_manager.update()
        entries = feed_manager.feed_entries
        assert entries is not None
        assert len(entries) == 5
        assert feed_manager.last_update is not None
        assert feed_manager.last_timestamp is None

        assert len(generated_entity_external_ids) == 5
        assert len(updated_entity_external_ids) == 0
        assert len(removed_entity_external_ids) == 0

        feed_entry = entries.get("3456")
        assert feed_entry.title == "Title 1"
        assert feed_entry.external_id == "3456"
        assert feed_entry.coordinates == (-37.2345, 149.1234)
        assert round(abs(feed_entry.distance_to_home - 714.4), 1) == 0
        assert repr(feed_entry) == "<MockFeedEntry(id=3456)>"

        feed_entry = entries.get("4567")
        assert feed_entry.title == "Title 2"
        assert feed_entry.external_id == "4567"

        feed_entry = entries.get("Title 3")
        assert feed_entry.title == "Title 3"
        assert feed_entry.external_id == "Title 3"

        external_id = hash((-37.8901, 149.7890))
        feed_entry = entries.get(external_id)
        assert feed_entry.title is None
        assert feed_entry.external_id == external_id

        feed_entry = entries.get("7890")
        assert feed_entry.title == "Title 5"
        assert feed_entry.external_id == "7890"

        # Simulate an update with several changes.
        generated_entity_external_ids.clear()
        updated_entity_external_ids.clear()
        removed_entity_external_ids.clear()

        mock_aioresponse.get(
            "http://test.url/testpath",
            status=HTTPStatus.OK,
            body=load_fixture("generic_feed_2.json"),
        )

        await feed_manager.update()
        entries = feed_manager.feed_entries
        assert entries is not None
        assert len(entries) == 3
        assert len(generated_entity_external_ids) == 1
        assert len(updated_entity_external_ids) == 2
        assert len(removed_entity_external_ids) == 3

        feed_entry = entries.get("3456")
        assert feed_entry.title == "Title 1 UPDATED"

        feed_entry = entries.get("4567")
        assert feed_entry.title == "Title 2"

        feed_entry = entries.get("8901")
        assert feed_entry.title == "Title 6"

        # Simulate an update with no data.
        generated_entity_external_ids.clear()
        updated_entity_external_ids.clear()
        removed_entity_external_ids.clear()

        with async_mock.patch(
            "aio_geojson_client.feed.GeoJsonFeed._fetch",
            new_callable=async_mock.AsyncMock,
        ) as mock_fetch:
            mock_fetch.return_value = (UPDATE_OK_NO_DATA, None)

            await feed_manager.update()
            entries = feed_manager.feed_entries

            assert len(entries) == 3
            assert len(generated_entity_external_ids) == 0
            assert len(updated_entity_external_ids) == 0
            assert len(removed_entity_external_ids) == 0

        # Simulate an update producing an error.
        generated_entity_external_ids.clear()
        updated_entity_external_ids.clear()
        removed_entity_external_ids.clear()

        mock_aioresponse.get(
            "http://test.url/testpath", status=HTTPStatus.INTERNAL_SERVER_ERROR
        )

        await feed_manager.update()
        entries = feed_manager.feed_entries

        assert len(entries) == 0
        assert len(generated_entity_external_ids) == 0
        assert len(updated_entity_external_ids) == 0
        assert len(removed_entity_external_ids) == 3

        # Simulate an update with dynamic filter.
        generated_entity_external_ids.clear()
        updated_entity_external_ids.clear()
        removed_entity_external_ids.clear()

        mock_aioresponse.get(
            "http://test.url/testpath",
            status=HTTPStatus.OK,
            body=load_fixture("generic_feed_1.json"),
        )
        await feed_manager.update_override(
            filter_overrides=GeoJsonFeedFilterDefinition(radius=750.0)
        )
        entries = feed_manager.feed_entries
        assert entries is not None
        assert len(entries) == 2
        assert feed_manager.last_update is not None
        assert feed_manager.last_timestamp is None


@pytest.mark.asyncio
async def test_feed_manager_with_status_callback(mock_aioresponse):
    """Test the feed manager."""
    home_coordinates = (-31.0, 151.0)
    mock_aioresponse.get(
        "http://test.url/testpath",
        status=HTTPStatus.OK,
        body=load_fixture("generic_feed_1.json"),
    )

    async with aiohttp.ClientSession(loop=asyncio.get_running_loop()) as websession:
        feed = MockGeoJsonFeed(websession, home_coordinates, "http://test.url/testpath")

        # This will just record calls and keep track of external ids.
        generated_entity_external_ids = []
        updated_entity_external_ids = []
        removed_entity_external_ids = []
        status_update = []

        async def _generate_entity(external_id):
            """Generate new entity."""
            generated_entity_external_ids.append(external_id)

        async def _update_entity(external_id):
            """Update entity."""
            updated_entity_external_ids.append(external_id)

        async def _remove_entity(external_id):
            """Remove entity."""
            removed_entity_external_ids.append(external_id)

        async def _status(status_details):
            """Capture status update details."""
            status_update.append(status_details)

        feed_manager = FeedManagerBase(
            feed, _generate_entity, _update_entity, _remove_entity, _status
        )
        assert (
            repr(feed_manager) == "<FeedManagerBase(feed=<"
            "MockGeoJsonFeed(home=(-31.0, 151.0), "
            "url=http://test.url/testpath, "
            "radius=None)>)>"
        )
        await feed_manager.update()
        entries = feed_manager.feed_entries
        assert entries is not None
        assert len(entries) == 5
        assert feed_manager.last_update is not None
        assert feed_manager.last_timestamp is None

        assert len(generated_entity_external_ids) == 5
        assert len(updated_entity_external_ids) == 0
        assert len(removed_entity_external_ids) == 0

        assert status_update[0].status == "OK"
        assert status_update[0].last_update is not None
        last_update_successful = status_update[0].last_update_successful
        assert status_update[0].last_update == last_update_successful
        assert status_update[0].last_timestamp is None
        assert status_update[0].total == 5
        assert status_update[0].created == 5
        assert status_update[0].updated == 0
        assert status_update[0].removed == 0
        assert (
            repr(status_update[0]) == f"<StatusUpdate("
            f"OK@{status_update[0].last_update})>"
        )

        # Simulate an update with no data.
        generated_entity_external_ids.clear()
        updated_entity_external_ids.clear()
        removed_entity_external_ids.clear()
        status_update.clear()

        mock_aioresponse.get(
            "http://test.url/testpath", status=HTTPStatus.INTERNAL_SERVER_ERROR
        )

        await feed_manager.update()
        entries = feed_manager.feed_entries

        assert len(entries) == 0
        assert len(generated_entity_external_ids) == 0
        assert len(updated_entity_external_ids) == 0
        assert len(removed_entity_external_ids) == 5

        assert status_update[0].status == "ERROR"
        assert status_update[0].last_update is not None
        assert status_update[0].last_update_successful is not None
        assert status_update[0].last_update_successful == last_update_successful
        assert status_update[0].total == 0



"""Constants."""
from typing import TypeVar

from .feed_entry import FeedEntry
from .filter_definition import GeoJsonFeedFilterDefinition

DEFAULT_REQUEST_TIMEOUT = 10

UPDATE_OK = "OK"
UPDATE_OK_NO_DATA = "OK_NO_DATA"
UPDATE_ERROR = "ERROR"

T_FILTER_DEFINITION = TypeVar("T_FILTER_DEFINITION", bound=GeoJsonFeedFilterDefinition)
T_FEED_ENTRY = TypeVar("T_FEED_ENTRY", bound=FeedEntry)



"""GeoJSON Feed."""
from __future__ import annotations

import asyncio
import logging
from abc import ABC, abstractmethod
from collections.abc import Callable
from datetime import datetime
from typing import Generic

import aiohttp
import geojson
from aiohttp import ClientSession, client_exceptions
from geojson import Feature, FeatureCollection

from .consts import (
    DEFAULT_REQUEST_TIMEOUT,
    T_FEED_ENTRY,
    T_FILTER_DEFINITION,
    UPDATE_ERROR,
    UPDATE_OK,
    UPDATE_OK_NO_DATA,
)

_LOGGER = logging.getLogger(__name__)


class GeoJsonFeed(Generic[T_FEED_ENTRY], ABC):
    """Geo JSON feed base class."""

    def __init__(
        self,
        websession: ClientSession,
        home_coordinates: tuple[float, float],
        url: str,
        filter_radius: float = None,
    ):
        """Initialise this service."""
        self._websession = websession
        self._home_coordinates = home_coordinates
        self._filter_radius = filter_radius
        self._url = url
        self._last_timestamp = None

    def __repr__(self):
        """Return string representation of this feed."""
        return "<{}(home={}, url={}, radius={})>".format(
            self.__class__.__name__,
            self._home_coordinates,
            self._url,
            self._filter_radius,
        )

    @abstractmethod
    def _new_entry(
        self, home_coordinates: tuple[float, float], feature, global_data: dict
    ) -> T_FEED_ENTRY:
        """Generate a new entry."""
        pass

    def _client_session_timeout(self) -> int:
        """Define client session timeout in seconds. Override if necessary."""
        return DEFAULT_REQUEST_TIMEOUT

    async def _update_internal(
        self, filter_function: Callable[[list[T_FEED_ENTRY]], list[T_FEED_ENTRY]]
    ) -> tuple[str, list[T_FEED_ENTRY] | None]:
        """Update from external source and return filtered entries."""
        status, data = await self._fetch()
        if status == UPDATE_OK:
            if data:
                entries = []
                global_data = self._extract_from_feed(data)
                # Extract data from feed entries.
                if type(data) is Feature:
                    entries.append(
                        self._new_entry(self._home_coordinates, data, global_data)
                    )
                elif type(data) is FeatureCollection:
                    for feature in data.features:
                        entries.append(
                            self._new_entry(
                                self._home_coordinates, feature, global_data
                            )
                        )
                else:
                    _LOGGER.warning(f"Unsupported GeoJSON object found: {type(data)}")
                filtered_entries = filter_function(entries)
                self._last_timestamp = self._extract_last_timestamp(filtered_entries)
                return UPDATE_OK, filtered_entries
            else:
                # Should not happen.
                return UPDATE_OK, None
        elif status == UPDATE_OK_NO_DATA:
            # Happens for example if the server returns 304
            return UPDATE_OK_NO_DATA, None
        else:
            # Error happened while fetching the feed.
            self._last_timestamp = None
            return UPDATE_ERROR, None

    async def update(self) -> tuple[str, list[T_FEED_ENTRY] | None]:
        """Update from external source and return filtered entries."""
        return await self._update_internal(
            lambda entries: self._filter_entries(entries)
        )

    async def update_override(
        self, filter_overrides: T_FILTER_DEFINITION = None
    ) -> tuple[str, list[T_FEED_ENTRY] | None]:
        """Update from external source and return filtered entries with ability to override filter conditions."""
        return await self._update_internal(
            lambda entries: self._filter_entries_override(
                entries, filter_overrides=filter_overrides
            )
        )

    async def _fetch(
        self, method: str = "GET", headers=None, params=None
    ) -> tuple[str, FeatureCollection | None]:
        """Fetch GeoJSON data from external source."""
        try:
            timeout = aiohttp.ClientTimeout(total=self._client_session_timeout())
            async with self._websession.request(
                method, self._url, headers=headers, params=params, timeout=timeout
            ) as response:
                try:
                    response.raise_for_status()
                    text = await response.text()
                    feature_collection = geojson.loads(text)
                    return UPDATE_OK, feature_collection
                except client_exceptions.ClientError as client_error:
                    _LOGGER.warning(
                        "Fetching data from %s failed with %s", self._url, client_error
                    )
                    return UPDATE_ERROR, None
                except ValueError as value_ex:
                    _LOGGER.warning(
                        "Unable to parse JSON from %s: %s", self._url, value_ex
                    )
                    return UPDATE_ERROR, None
        except client_exceptions.ClientError as client_error:
            _LOGGER.warning(
                "Requesting data from %s failed with " "client error: %s",
                self._url,
                client_error,
            )
            return UPDATE_ERROR, None
        except asyncio.TimeoutError:
            _LOGGER.warning(
                "Requesting data from %s failed with " "timeout error", self._url
            )
            return UPDATE_ERROR, None

    def _filter_entries(self, entries: list[T_FEED_ENTRY]) -> list[T_FEED_ENTRY]:
        """Filter the provided entries (for backwards-compatibility)."""
        return self._filter_entries_override(entries, None)

    def _filter_entries_override(
        self, entries: list[T_FEED_ENTRY], filter_overrides: T_FILTER_DEFINITION = None
    ) -> list[T_FEED_ENTRY]:
        """Filter the provided entries with ability to override filter definitions."""
        filtered_entries = entries
        _LOGGER.debug("Entries before filtering %s", filtered_entries)
        # Always remove entries without geometry
        filtered_entries = list(
            filter(
                lambda entry: entry.geometries is not None
                and len(entry.geometries) >= 1,
                filtered_entries,
            )
        )
        # Filter by distance.
        filter_radius = (
            filter_overrides.radius
            if filter_overrides and filter_overrides.radius
            else self._filter_radius
        )
        if filter_radius:
            filtered_entries = list(
                filter(
                    lambda entry: entry.distance_to_home <= filter_radius,
                    filtered_entries,
                )
            )
        _LOGGER.debug("Entries after filtering %s", filtered_entries)
        return filtered_entries

    @abstractmethod
    def _extract_from_feed(self, feed: FeatureCollection) -> dict | None:
        """Extract global metadata from feed."""
        return None

    @abstractmethod
    def _extract_last_timestamp(
        self, feed_entries: list[T_FEED_ENTRY]
    ) -> datetime | None:
        """Determine latest (newest) entry from the filtered feed."""
        return None

    @property
    def last_timestamp(self) -> datetime | None:
        """Return the last timestamp extracted from this feed."""
        return self._last_timestamp



"""Filter definitions."""


class GeoJsonFeedFilterDefinition:
    """Filter definition."""

    def __init__(self, radius: float = None):
        """Initialise filter definition."""
        self._radius = radius

    @property
    def radius(self) -> float:
        """Return the radius."""
        return self._radius

    @radius.setter
    def radius(self, value: float):
        """Set radius."""
        self._radius = value



"""GeoJSON point."""
from __future__ import annotations

from .geometry import Geometry


class Point(Geometry):
    """Represents a point."""

    def __init__(self, latitude: float, longitude: float):
        """Initialise point."""
        self._latitude = latitude
        self._longitude = longitude

    def __repr__(self):
        """Return string representation of this point."""
        return "<{}(latitude={}, longitude={})>".format(
            self.__class__.__name__, self.latitude, self.longitude
        )

    def __hash__(self) -> int:
        """Return unique hash of this point."""
        return hash((self.latitude, self.longitude))

    def __eq__(self, other: object) -> bool:
        """Return if this object is equal to other object."""
        return (
            self.__class__ == other.__class__
            and self.latitude == other.latitude
            and self.longitude == other.longitude
        )

    @property
    def latitude(self) -> float | None:
        """Return the latitude of this point."""
        return self._latitude

    @property
    def longitude(self) -> float | None:
        """Return the longitude of this point."""
        return self._longitude



"""Test for the generic geojson feed."""
import asyncio
from http import HTTPStatus
from unittest.mock import MagicMock

import aiohttp
import pytest
from aiohttp import ClientOSError

from aio_geojson_client.consts import UPDATE_ERROR, UPDATE_OK
from aio_geojson_client.filter_definition import GeoJsonFeedFilterDefinition
from aio_geojson_client.geometries.point import Point
from aio_geojson_client.geometries.polygon import Polygon
from tests import MockGeoJsonFeed
from tests.utils import load_fixture


@pytest.mark.asyncio
async def test_update_ok(mock_aioresponse):
    """Test updating feed is ok."""
    home_coordinates = (-31.0, 151.0)
    mock_aioresponse.get(
        "http://test.url/testpath",
        status=HTTPStatus.OK,
        body=load_fixture("generic_feed_1.json"),
    )

    async with aiohttp.ClientSession(loop=asyncio.get_running_loop()) as websession:
        feed = MockGeoJsonFeed(websession, home_coordinates, "http://test.url/testpath")
        assert (
            repr(feed) == "<MockGeoJsonFeed(home=(-31.0, 151.0), "
            "url=http://test.url/testpath, radius=None)>"
        )
        status, entries = await feed.update()
        assert status == UPDATE_OK
        assert entries is not None
        assert len(entries) == 5

        feed_entry = entries[0]
        assert feed_entry is not None
        assert feed_entry.title == "Title 1"
        assert feed_entry.external_id == "3456"
        assert feed_entry.coordinates == (-37.2345, 149.1234)
        assert round(abs(feed_entry.distance_to_home - 714.4), 1) == 0
        assert repr(feed_entry) == "<MockFeedEntry(id=3456)>"

        feed_entry = entries[1]
        assert feed_entry is not None
        assert feed_entry.title == "Title 2"
        assert feed_entry.external_id == "4567"

        feed_entry = entries[2]
        assert feed_entry is not None
        assert feed_entry.title == "Title 3"
        assert feed_entry.external_id == "Title 3"

        feed_entry = entries[3]
        assert feed_entry is not None
        assert feed_entry.title is None
        assert feed_entry.external_id == hash(feed_entry.coordinates)

        feed_entry = entries[4]
        assert feed_entry is not None
        assert feed_entry.title == "Title 5"
        assert feed_entry.external_id == "7890"


@pytest.mark.asyncio
async def test_update_ok_with_filtering(mock_aioresponse):
    """Test updating feed is ok."""
    home_coordinates = (-37.0, 150.0)
    mock_aioresponse.get(
        "http://test.url/testpath",
        status=HTTPStatus.OK,
        body=load_fixture("generic_feed_1.json"),
    )

    async with aiohttp.ClientSession(loop=asyncio.get_running_loop()) as websession:
        feed = MockGeoJsonFeed(
            websession, home_coordinates, "http://test.url/testpath", filter_radius=90.0
        )
        status, entries = await feed.update()
        assert status == UPDATE_OK
        assert entries is not None
        assert len(entries) == 4
        assert round(abs(entries[0].distance_to_home - 82.0), 1) == 0
        assert round(abs(entries[1].distance_to_home - 77.0), 1) == 0
        assert round(abs(entries[2].distance_to_home - 84.6), 1) == 0


@pytest.mark.asyncio
async def test_update_ok_with_filter_override(mock_aioresponse):
    """Test updating feed is ok."""
    home_coordinates = (-37.0, 150.0)
    mock_aioresponse.get(
        "http://test.url/testpath",
        status=HTTPStatus.OK,
        body=load_fixture("generic_feed_1.json"),
    )

    async with aiohttp.ClientSession(loop=asyncio.get_running_loop()) as websession:
        feed = MockGeoJsonFeed(
            websession, home_coordinates, "http://test.url/testpath", filter_radius=60.0
        )
        status, entries = await feed.update_override(
            filter_overrides=GeoJsonFeedFilterDefinition(radius=90.0)
        )
        assert status == UPDATE_OK
        assert entries is not None
        assert len(entries) == 4
        assert round(abs(entries[0].distance_to_home - 82.0), 1) == 0
        assert round(abs(entries[1].distance_to_home - 77.0), 1) == 0
        assert round(abs(entries[2].distance_to_home - 84.6), 1) == 0


@pytest.mark.asyncio
async def test_update_geometries(mock_aioresponse):
    """Test updating feed is ok."""
    home_coordinates = (-31.0, 151.0)
    mock_aioresponse.get(
        "http://test.url/testpath",
        status=HTTPStatus.OK,
        body=load_fixture("generic_feed_3.json"),
    )

    async with aiohttp.ClientSession(loop=asyncio.get_running_loop()) as websession:
        feed = MockGeoJsonFeed(websession, home_coordinates, "http://test.url/testpath")
        assert (
            repr(feed) == "<MockGeoJsonFeed(home=(-31.0, 151.0), "
            "url=http://test.url/testpath, radius=None)>"
        )
        status, entries = await feed.update()
        assert status == UPDATE_OK
        assert entries is not None
        assert len(entries) == 3

        feed_entry = entries[0]
        assert feed_entry is not None
        assert feed_entry.title == "Title 1"
        assert feed_entry.external_id == "1234"
        assert isinstance(feed_entry.geometries[0], Point)
        assert feed_entry.coordinates == (-35.5, 150.5)
        assert round(abs(feed_entry.distance_to_home - 502.5), 1) == 0
        assert repr(feed_entry) == "<MockFeedEntry(id=1234)>"

        feed_entry = entries[1]
        assert feed_entry is not None
        assert feed_entry.title == "Title 2"
        assert feed_entry.external_id == "2345"
        assert isinstance(feed_entry.geometries[0], Polygon)
        assert feed_entry.coordinates == (-28.0, 152.0)

        feed_entry = entries[2]
        assert feed_entry is not None
        assert feed_entry.title == "Title 3"
        assert feed_entry.external_id == "3456"
        assert isinstance(feed_entry.geometries[0], Point)
        assert isinstance(feed_entry.geometries[1], Polygon)
        assert feed_entry.coordinates == (-35.5, 150.5)


@pytest.mark.asyncio
async def test_update_with_client_exception():
    """Test updating feed results in error."""
    home_coordinates = (-31.0, 151.0)

    async with aiohttp.ClientSession(loop=asyncio.get_running_loop()):
        mock_websession = MagicMock()
        mock_websession.request.side_effect = ClientOSError
        feed = MockGeoJsonFeed(
            mock_websession, home_coordinates, "http://test.url/badpath"
        )
        status, entries = await feed.update()
        assert status == UPDATE_ERROR
        assert feed.last_timestamp is None


@pytest.mark.asyncio
async def test_update_with_request_exception(mock_aioresponse):
    """Test updating feed results in error."""
    home_coordinates = (-31.0, 151.0)
    mock_aioresponse.get("http://test.url/testpath", status=HTTPStatus.NOT_FOUND)

    async with aiohttp.ClientSession(loop=asyncio.get_running_loop()) as websession:
        feed = MockGeoJsonFeed(websession, home_coordinates, "http://test.url/badpath")
        status, entries = await feed.update()
        assert status == UPDATE_ERROR
        assert feed.last_timestamp is None


@pytest.mark.asyncio
async def test_update_with_json_decode_error(mock_aioresponse):
    """Test updating feed raises exception."""
    home_coordinates = (-31.0, 151.0)
    mock_aioresponse.get(
        "http://test.url/badjson", status=HTTPStatus.OK, body="NOT JSON"
    )

    async with aiohttp.ClientSession(loop=asyncio.get_running_loop()) as websession:
        feed = MockGeoJsonFeed(websession, home_coordinates, "http://test.url/badjson")
        status, entries = await feed.update()
        assert status == UPDATE_ERROR
        assert entries is None


@pytest.mark.asyncio
async def test_update_with_timeout_error():
    """Test updating feed results in timeout error."""
    home_coordinates = (-31.0, 151.0)

    async with aiohttp.ClientSession(loop=asyncio.get_running_loop()):
        mock_websession = MagicMock()
        mock_websession.request.side_effect = asyncio.TimeoutError
        feed = MockGeoJsonFeed(
            mock_websession, home_coordinates, "http://test.url/goodpath"
        )
        status, entries = await feed.update()
        assert status == UPDATE_ERROR
        assert feed.last_timestamp is None


@pytest.mark.asyncio
async def test_update_ok_feed_feature(mock_aioresponse):
    """Test updating feed is ok."""
    home_coordinates = (-31.0, 151.0)
    mock_aioresponse.get(
        "http://test.url/testpath",
        status=HTTPStatus.OK,
        body=load_fixture("generic_feed_4.json"),
    )

    async with aiohttp.ClientSession(loop=asyncio.get_running_loop()) as websession:
        feed = MockGeoJsonFeed(websession, home_coordinates, "http://test.url/testpath")
        assert (
            repr(feed) == "<MockGeoJsonFeed(home=(-31.0, 151.0), "
            "url=http://test.url/testpath, radius=None)>"
        )
        status, entries = await feed.update()
        assert status == UPDATE_OK
        assert entries is not None
        assert len(entries) == 1

        feed_entry = entries[0]
        assert feed_entry is not None
        assert feed_entry.title == "Title 1"
        assert feed_entry.external_id == "3456"
        assert feed_entry.coordinates == (-37.2345, 149.1234)
        assert round(abs(feed_entry.distance_to_home - 714.4), 1) == 0
        assert repr(feed_entry) == "<MockFeedEntry(id=3456)>"


@pytest.mark.asyncio
async def test_unsupported_object(mock_aioresponse, caplog):
    """Test updating feed is ok."""
    home_coordinates = (-31.0, 151.0)
    mock_aioresponse.get(
        "http://test.url/testpath",
        status=HTTPStatus.OK,
        body=load_fixture("generic_feed_5.json"),
    )

    async with aiohttp.ClientSession(loop=asyncio.get_running_loop()) as websession:
        feed = MockGeoJsonFeed(websession, home_coordinates, "http://test.url/testpath")
        assert (
            repr(feed) == "<MockGeoJsonFeed(home=(-31.0, 151.0), "
            "url=http://test.url/testpath, radius=None)>"
        )
        status, entries = await feed.update()
        assert status == UPDATE_OK
        assert entries is not None
        assert len(entries) == 0

        assert (
            "Unsupported GeoJSON object found: <class 'geojson.geometry.Point'>"
            in caplog.text
        )



"""Tests for base classes."""
import unittest
from unittest.mock import ANY, MagicMock

from aio_geojson_client.geojson_distance_helper import GeoJsonDistanceHelper
from aio_geojson_client.geometries.point import Point
from aio_geojson_client.geometries.polygon import Polygon


class TestGeoJsonDistanceHelper(unittest.TestCase):
    """Tests for the GeoJSON distance helper."""

    def test_extract_coordinates_from_point(self):
        """Test extracting coordinates from point."""
        point = Point(-30.0, 151.0)
        latitude, longitude = GeoJsonDistanceHelper.extract_coordinates(point)
        assert latitude == -30.0
        assert longitude == 151.0

    def test_extract_coordinates_from_polygon(self):
        """Test extracting coordinates from polygon."""
        polygon = Polygon(
            [
                Point(-30.0, 151.0),
                Point(-30.0, 151.5),
                Point(-30.5, 151.5),
                Point(-30.5, 151.0),
                Point(-30.0, 151.0),
            ]
        )
        latitude, longitude = GeoJsonDistanceHelper.extract_coordinates(polygon)
        self.assertAlmostEqual(latitude, -30.2, 1)
        self.assertAlmostEqual(longitude, 151.2, 1)

    def test_extract_coordinates_from_unsupported_geometry(self):
        """Test extracting coordinates from unsupported geometry."""
        mock_unsupported_geometry = MagicMock()
        latitude, longitude = GeoJsonDistanceHelper.extract_coordinates(
            mock_unsupported_geometry
        )
        self.assertIsNone(latitude)
        self.assertIsNone(longitude)

    def test_distance_to_point(self):
        """Test calculating distance to point."""
        home_coordinates = (-31.0, 150.0)
        point = Point(-30.0, 151.0)
        distance = GeoJsonDistanceHelper.distance_to_geometry(home_coordinates, point)
        self.assertAlmostEqual(distance, 146.8, 1)

    def test_distance_to_polygon_1(self):
        """Test calculating distance to point."""
        home_coordinates = (-31.0, 150.0)
        polygon = Polygon(
            [
                Point(-30.0, 151.0),
                Point(-30.0, 151.5),
                Point(-30.5, 151.5),
                Point(-30.5, 151.0),
                Point(-30.0, 151.0),
            ]
        )
        distance = GeoJsonDistanceHelper.distance_to_geometry(home_coordinates, polygon)
        self.assertAlmostEqual(distance, 110.6, 1)

    def test_distance_to_polygon_2(self):
        """Test calculating distance to polygon."""
        home_coordinates = (-30.2, 151.2)
        polygon = Polygon(
            [
                Point(-30.0, 151.0),
                Point(-30.0, 151.5),
                Point(-30.5, 151.5),
                Point(-30.5, 151.0),
                Point(-30.0, 151.0),
            ]
        )
        distance = GeoJsonDistanceHelper.distance_to_geometry(home_coordinates, polygon)
        self.assertAlmostEqual(distance, 0.0, 1)

    def test_distance_to_polygon_3(self):
        """Test calculating distance to polygon."""
        home_coordinates = (-29.0, 151.2)
        polygon = Polygon(
            [
                Point(-30.0, 151.0),
                Point(-30.0, 151.5),
                Point(-30.5, 151.5),
                Point(-30.5, 151.0),
                Point(-30.0, 151.0),
            ]
        )
        distance = GeoJsonDistanceHelper.distance_to_geometry(home_coordinates, polygon)
        self.assertAlmostEqual(distance, 111.2, 1)

    def test_distance_to_polygon_4(self):
        """Test calculating distance to polygon."""
        home_coordinates = (30.0, 151.3)
        polygon = Polygon(
            [
                Point(30.0, 151.0),
                Point(30.0, 151.5),
                Point(30.5, 151.5),
                Point(30.5, 151.0),
                Point(30.0, 151.0),
            ]
        )
        distance = GeoJsonDistanceHelper.distance_to_geometry(home_coordinates, polygon)
        self.assertAlmostEqual(distance, 0.0, 1)

    def test_distance_to_polygon_5(self):
        """Test calculating distance to polygon."""
        polygon = Polygon(
            [
                Point(30.0, 179.0),
                Point(30.0, -179.5),
                Point(30.5, -179.5),
                Point(30.5, 179.0),
                Point(30.0, 179.0),
            ]
        )
        home_coordinates = (30.2, -177.0)
        distance = GeoJsonDistanceHelper.distance_to_geometry(home_coordinates, polygon)
        self.assertAlmostEqual(distance, 240.3, 1)
        home_coordinates = (30.1, 178.0)
        distance = GeoJsonDistanceHelper.distance_to_geometry(home_coordinates, polygon)
        self.assertAlmostEqual(distance, 96.2, 1)
        home_coordinates = (31.0, -179.8)
        distance = GeoJsonDistanceHelper.distance_to_geometry(home_coordinates, polygon)
        self.assertAlmostEqual(distance, 55.6, 1)
        home_coordinates = (31.0, 179.8)
        distance = GeoJsonDistanceHelper.distance_to_geometry(home_coordinates, polygon)
        self.assertAlmostEqual(distance, 55.6, 1)

    def test_distance_to_polygon_6(self):
        """Test calculating distance to polygon."""
        polygon = Polygon(
            [
                Point(-30.0, 179.0),
                Point(-30.0, -179.5),
                Point(-29.5, -179.5),
                Point(-29.5, 179.0),
                Point(-30.0, 179.0),
            ]
        )
        home_coordinates = (-29.8, -177.0)
        distance = GeoJsonDistanceHelper.distance_to_geometry(home_coordinates, polygon)
        self.assertAlmostEqual(distance, 241.2, 1)
        home_coordinates = (-29.9, 178.0)
        distance = GeoJsonDistanceHelper.distance_to_geometry(home_coordinates, polygon)
        self.assertAlmostEqual(distance, 96.4, 1)
        home_coordinates = (-29.0, -179.8)
        distance = GeoJsonDistanceHelper.distance_to_geometry(home_coordinates, polygon)
        self.assertAlmostEqual(distance, 55.6, 1)
        home_coordinates = (-29.0, 179.8)
        distance = GeoJsonDistanceHelper.distance_to_geometry(home_coordinates, polygon)
        self.assertAlmostEqual(distance, 55.6, 1)

    def test_distance_to_unsupported_geometry(self):
        """Test calculating distance to unsupported geometry."""
        home_coordinates = (-31.0, 150.0)
        mock_unsupported_geometry = MagicMock()
        distance = GeoJsonDistanceHelper.distance_to_geometry(
            home_coordinates, mock_unsupported_geometry
        )
        assert distance == float("inf")

    def test_perpendicular_point_edge_case(self):
        """Test edge case when calculating perpendicular point."""
        edge = (Point(-31.0, 150.0), Point(-31.0, 150.0))
        result = GeoJsonDistanceHelper._perpendicular_point(edge, ANY)
        assert result is None
